package RFC::Multiplex;
#$Id: Multiplex.pm,v 1.91 2016/11/22 07:56:05 sqko Exp $

###################################################################
#
#
#     RFC::Multiplex
# 
#  This class defines the control of multiple gasses to the same
#  MFC. To obtain a Multiplexer instance call the constructor:
#
#  $id = RFC::Multiplex->new($name,$rig)
#
#  The constructor can also be called with an additional argument
#  which must then be a RFC::MFC (or derived) device instance:
#
#  $mfc = RFC::MFC->new($name,$rig);
#
#  $id = RFC::Multiplex->new($name,$rig,$mfc);
#
#  
#  All Multiplex objects have the following member functions (excluding 
#  those derived from RFC::BaseDevce and Debug.pm):
#
#  $id->attatch_controler($)   Attatches an RFC::MFC instance to the
#                              multiplexer (if not included in the
#                              call of the constructor).
#  $id->set($)                 Sets the multiplexer to select the
#                              specified gas name.
#  $id->gasses()               Returns a list of possible gasses.
#  $id->gas()                  Returns the name of the selected gas (NB NOT device name).
#                              If supplied with an optional argument this is used
#                              as a data structure to access if the device had
#                              already been read (similar to the read() and readstring())
#                              member functions inherited rom BaseDevice.
#  $id->gas_name()             Returns the device name of the selected gas.
#                              If supplied with an optional argument this is used
#                              as a data structure to access if the device had
#                              already been read (similar to the read() and readstring())
#                              member functions inherited rom BaseDevice.
#
# Prior to version 4.2.5, the implementation of the RFC::Multiplex class assumed 
# that all gasses were controled from the same relay board (usually a ICP-con relay module)
# Thus all gas relays shared the same device tty and address, but differed
# only in the channel number. As of version 5.0 this is deprecated and will only
# be avaliable for configuration if legacy mode is enabled (it will still work for
# already configured systems).
#
# As of version 4.2.5 and onwards, it is possible to use arbitrary relay
# devices to control a gas multiplexer. To do so, simply configure the relays
# as normal RFC::Relay instances and select the proper names (defined in the
# '$gasname'_device_name key in the configuration section ($gasname being the name of
# the gas line in question).
#
#
###################################################################

use strict;
use warnings;
use RFC::Header;
use SemaforeFile;
use RFC::BaseDevice;
use RFC::Cache;
our @ISA = qw(RFC::BaseDevice);

BEGIN {
    use Exporter ();
    our ($VERSION, @ISA, @EXPORT, @EXPORT_OK, %EXPORT_TAGS);
    $VERSION = 1.0;
    %EXPORT_TAGS = (
        'standard'=>[],
                    );
    Exporter::export_ok_tags('standard');
    Exporter::export_tags('standard');
    @ISA = qw(Exporter);
}

sub new()
{
    my $class = shift;
    my $name = shift;
    my $rig = shift;
    my $self = $class->SUPER::new($name,$rig);
    $self->{'_controler'} = undef;
    $self->{'_controler'} = shift if (@_);
    $self->{'_setup_tags'} = ['type','mode','title','show_plot','use_cache'];
    $self->{'_legacy_tags'} = ['type','mode','tty','address','control_type','title','show_plot','use_cache'];
    $self->{'_desc'} = {
        'mode' => 'Device control mode',
        'tty' => 'Device communication device (ex. ttyS0)',
        'address' => 'Device address',
	'control_type' => 'Device control type',
	'title' => 'Optional title',
        'use_cache'=>'Determines if persisten caching of read values is allowed',
    };
    $self->{'_default'} = {
        'mode' => 'Automatic',
        'tty' => 'ttyS0',
        'address' => 0,
	'control_type' => 'ICP',
        'use_cache'=>'No',
    };
    $self->{'_values_type'} = [(RFC::Device::list_Multiplexer())];
    $self->{'_values_control_type'} = [(RFC::Device::list_RELAY())];
    $self->{'_section'} = 'gas_multiplexer_'.$self->name;
    bless $self,$class;
    $self->load_conf();
    $self->{'_type'} = 'Relay';
    $self->{'_file'} = SemaforeFile->new($userdir.'rig'.$self->rig->name.'/multiplex_'.$self->name,'/home/celltest/Gas_rig'.$self->rig->name.'.lock');
    unless ($self->{'_file'}->exist)
    {
        ### Create file with rw for all           ###
        ### Yes I know 666 is the number of evil! ###
        $self->{'_file'}->lock_ex;
        $self->{'_file'}->writeline('');
        chmod 0666, $self->{'_file'}->filename;
        $self->{'_file'}->unlock;
    }
    $self->{'_info_string'} = 'Note that gas multiplexers are not normally required to be enabled to work properly and be controlable from the user interface as they are enabled through the MFC device they are attatched to';
    return $self->init;
}

sub DESTROY
{
    my $self = shift;
    $self->{'_file'}->DESTROY() if ($self->{'_file'});
    return undef;
}


sub attatch_controler($)
{
    my $self = shift;
    $self->{'_controler'} = shift;
    return $self->init;
}

sub init
{
    my $self = shift;
    $self->{'_gasses'} = [];
    $self->{'_channels'} = {};
    $self->{'_gasses'} = [$self->{'_controler'}->gasses] if ($self->{'_controler'});
    if ($self->mode =~ /auto/i)
    {
        my $type = $self->get_conf('control_type');
        $type = 'ICP' unless $type;
	foreach my $gas (@{$self->{'_gasses'}})
	{
	    my $dev = $self->rig->get_device('relay',$self->get_conf($gas.'_device_name'));
	    unless ($dev)
	    {
		$dev = $self->rig->init_device('relay',$self->get_conf($gas.'_device_name'));
		$dev->readstring_ignore(1) if ($dev);
	    }
	    unless ($dev)
	    {
		$dev = RFC::Device::new_RELAY($self->rig,'multiplex_'.$self->name.'_'.$gas,$type,$self->{'_tty'},$self->{'_address'},$self->channel($gas));
		$dev->readstring_ignore(1) if ($dev);
	    }
	    $self->die("Eerror: Could not initialise relay device for gas $gas") unless ($dev);
	    $self->{'_channels'}{$gas} = $dev;
	}
    }
    else
    {
	foreach my $gas (@{$self->{'_gasses'}})
	{
	    my $dev = RFC::Device::new_RELAY($self->rig,'multiplex_'.$self->name.'_'.$gas,'Manual');
	    $dev->readstring_ignore(1);
	    $self->{'_channels'}{$gas} = $dev;
	}
    }
    return $self;
}

sub test()
{
    my $self = shift;
    my $ok = 1;
    foreach my $g (@{$self->{'_gasses'}})
    {
	$ok = 0 unless ($self->{'_channels'}{$g}->test);
    }
    return $ok;
}


sub Accept
{
    my $self = shift;
    my $visitor= shift;
    return undef if ($self->{'_seen_visitors'}{$visitor});
    foreach my $g (@{$self->{'_gasses'}})
    {
	$self->{'_channels'}{$g}->Accept($visitor) if ($self->{'_channels'}{$g});
    }
    $visitor->VisitComplex($self);
    $self->{'_seen_visitors'}{$visitor} = 1;
    return undef;
}


sub test_string()
{
    my $self = shift;
    my $string = 'Multiplexer '.$self->name.': ';
    my @arr = ();
    my $data = RFC::Cache->new();
    foreach my $g (@{$self->{'_gasses'}})
    {
	push @arr,$g.' => '.$self->{'_channels'}{$g}->read($data);
    }
    $string .= join(', ',@arr)."\n";
    return $string;
}


sub monitor()
{
    my $self = shift;
    my $data = undef;
    if (@_)
    {
        $data = $_[0];
    }
    my @arr = ();
    foreach my $g (@{$self->{'_gasses'}})
    {
	push @arr,$g.' => '.$self->{'_channels'}{$g}->read($data);
    }
    return join(', ',@arr)."\n";
}

sub gasses()
{
    my $self = shift;
    $self->init() if (!@{$self->{'_gasses'}} && $self->{'_controler'});
    return @{$self->{'_gasses'}};
}

sub gas_all()
{
    my $self = shift;
    my $data = undef;
    if (@_)
    {
        $data = $_[0];
    }
    else
    {
        $data = RFC::Cache->new();
        $data->debug($self->debug()) if ($self->debug);
    }
    $self->check_cache($data);
    my @gas = ();
    foreach my $g (@{$self->{'_gasses'}})
    {
	my $status = $self->{'_channels'}{$g}->read($data);
	next unless ($status == 1);  ## if read returns -1 it is an error value ##
	my $gdev = $self->rig->get_device('gas',$g);
	push @gas, $gdev->gas() if ($gdev);
    }
    unless (@gas == 1)
    {
	$self->writedebug("Retrying as no gas appears to be selected!");
	@gas = ();
	## Retry if communication failed ##
	my $d2 = RFC::Cache->new();
	foreach my $g (@{$self->{'_gasses'}})
	{
	    my $status = $self->{'_channels'}{$g}->read($d2);
	    next unless ($status == 1);   ## if read returns -1 it is an error value ##
	    my $gdev = $self->rig->get_device('gas',$g);
	    push @gas, $gdev->gas() if ($gdev);
	}
    }
    unless (@gas)
    {
	my $str = 'Warning: No gas was engaged for multiplexer '.$self->name.' for rig '.$self->rig->name;
        $self->rig->warning_mail($self,$str);
	$self->writedebug($str);
	return undef;
    }
    if (@gas > 1)
    {
	my $str = 'Alert! More than one gas is selected for multiplexer '.$self->name.' for rig ' . $self->rig->name . '. Selected gasses: '.join(', ',@gas);
        $self->rig->warning_mail($self,$str);
	$self->writedebug($str);
	return undef;
    }
    return @gas;
}

sub gas()
{
    my $self = shift;
    my @list = $self->gas_all(@_);
    return $list[0];
}

sub gas_all_name()
{
    my $self = shift;
    my $data = undef;
    if (@_)
    {
        $data = $_[0];
    }
    else
    {
        $data = RFC::Cache->new();
        $data->debug($self->debug()) if ($self->debug);
    }
    $self->check_cache($data);
    my @gas = ();
    foreach my $g (@{$self->{'_gasses'}})
    {
	my $status = $self->{'_channels'}{$g}->read($data);
	next unless ($status == 1);  ## if read returns -1 it is an error value ##
	my $gdev = $self->rig->get_device('gas',$g);
	push @gas, $gdev->name() if ($gdev);
    }
    unless (@gas == 1)
    {
	$self->writedebug("Retrying as no gas appears to be selected!");
	@gas = ();
	## Retry if communication failed ##
	my $d2 = RFC::Cache->new();
	foreach my $g (@{$self->{'_gasses'}})
	{
	    my $status = $self->{'_channels'}{$g}->read($d2);
	    next unless ($status == 1);   ## if read returns -1 it is an error value ##
	    my $gdev = $self->rig->get_device('gas',$g);
	    push @gas, $gdev->name() if ($gdev);
	}
    }
    unless (@gas)
    {
	my $str = 'Warning: No gas was engaged for multiplexer '.$self->name.' for rig '.$self->rig->name;
        $self->rig->warning_mail($self,$str);
	$self->writedebug($str);
	return undef;
    }
    if (@gas > 1)
    {
	my $str = 'Alert! More than one gas is selected for multiplexer '.$self->name.' for rig ' . $self->rig->name . '. Selected gasses: '.join(', ',@gas);
        $self->rig->warning_mail($self,$str);
	$self->writedebug($str);
	return undef;
    }
    return @gas;
}

sub gas_name()
{
    my $self = shift;
    my @list = $self->gas_all_name(@_);
    return $list[0];
}

sub readstring()
{
    my $self = shift;
    return undef if ($self->readstring_ignore);
    my $data = undef;
    if (@_)
    {
        $data = $_[0];
    }
    else
    {
        $data = RFC::Cache->new();
        $data->debug($self->debug()) if ($self->debug);
    }
    $self->check_cache($data);
    my $res = undef;
    my @t = $self->get_time();
    my $t = join(':',@t);
    my @gas = $self->gas_all_name($data);
    my $gas = 'NONE';
    $gas = join(',',@gas) if (@gas);
    $gas = 'NA' unless ($gas && $gas =~ /\w/);
    return join(' ',$self->name.'_selected_gas',$t,time,$gas);
}

sub read_nocache
{
    my $self = shift;
    my $gas = shift;
    my $data = undef;
    if (@_)
    {
        $data = $_[0];
    }
    else
    {
        $data = RFC::Cache->new();
        $data->debug($self->debug()) if ($self->debug);
    }
    $self->check_cache($data);
    unless ($gas)
    {
	return $self->gas($data);
    }
    unless (exists($self->{'_channels'}{$gas}))
    {
	$self->writedebug("Warning, None of the gasses for multiplexer ".$self->name." was found matching gas $gas for rig ".$self->rig->name);
	return undef;
    }
    my $res = $self->{'_channels'}{$gas}->read($data);
    $self->writedebug("Status for multiplexer ".$self->name." for gas $gas on channel ".$self->channel($gas)." was $res") if ($self->debug > 2);
    $res = 0 unless (defined($res));
    my $str = time .' '.$self->name. ' '.$res;
    $self->rig->kanaldata($str) unless ($self->readstring_ignore);
    $self->set_cache($res);
    $self->{'_last_read'} = Time::HiRes::time;
    return $res;
}

sub set()
{
    my $self = shift;
    my $gas = shift;
    $self->writedebug("Setting multiplexer ".$self->name." to $gas") if ($self->debug > 2);
    unless (exists($self->{'_channels'}{$gas}))
    {
	$self->die("Error: None of the gasses for multiplexer ".$self->name." was found matching gas $gas for rig ".$self->rig->name);
	return undef;
    }
    my $data = RFC::Cache->new();
    foreach my $g (@{$self->{'_gasses'}})
    {
	my $status = $self->{'_channels'}{$g}->read($data);
	## Only set relays to 0 for engaged relays (saves serial bus bandwidth) ##
	next if ($status == 0 || $g eq $gas);
	$self->{'_channels'}{$g}->set(0,@_);
    }
    $self->{'_channels'}{$gas}->set(1,@_);
    $self->rig->change_config_value('gas_'.$gas,'mode','Automatic');
    ### Set gas flows for not engaged gasses to 0 and mark as manual ###
    $self->writedebug("Setting non engaged gas flows to 0") if ($self->debug > 2);
    foreach my $g (@{$self->{'_gasses'}})
    {
	next if ($g eq $gas);
	$self->rig->change_config_value('gas_'.$g,'mode','Manual');
	my $dev = $self->rig->get_device('gas',$g);
	if ($dev)
	{
	    $dev->setflow(0,@_);
	}
	else
	{
	    $self->errorlog("Error: could not get device for gas $gas when changing multiplexer ".$self->name,$_[0]);
	}
    }
    ### Set gas for controler ###
    unless ($self->{'_controler'})
    {
	foreach my $d ($self->rig->list_devices('MFC'))
	{
	    my $dev = $self->rig->get_device('MFC',$d);
	    $self->{'_controler'} = $d if ($dev && $dev->multiplexer == $self->name);
	}
    }
    $self->rig->change_config_value('controler_'.$self->{'_controler'}->name,'gas',$gas) if ($self->{'_controler'});
    ### Reset error lock ###
    $self->rig->unset_warning($self);
    $self->notify('set',$gas,@_);
}

sub print_config()
{
    my $self = shift;
    my @arr = ();
    push @arr,'Name: '.$self->name;
    push @arr,'Mode: '.$self->mode;
    push @arr,'Type: '.$self->type;
    push @arr,'TTY: '.$self->{'_tty'};
    push @arr,'Address: '.$self->{'_address'};
    my @chs = ();
    foreach my $g (@{$self->{'_gasses'}})
    {
	my $dev = $self->rig->init_device('relay',$self->get_conf($g.'_device_name'));
	if ($dev)
	{
	    push @chs,$g.'->'.$dev->print_config;
	}
	else
	{
	    push @chs,$g.'->'.$self->channel($g);
	}
    }    
    push @arr,'channels: '.join(',',@chs);
    return join (', ',@arr);
}


sub channel($)
{
    my $self = shift;
    my $gas = shift;
    my $ch = $self->get_conf($gas);
    return $ch;
}




package RFC::VICI;
our @ISA = qw(RFC::Multiplex);

sub new()
{
    my $class = shift;
    my $name = shift;
    my $rig = shift;
    my $self = $class->SUPER::new($name,$rig);
    $self->{'_controler'} = undef;
    $self->{'_controler'} = shift if (@_);
    $self->{'_setup_tags'} = ['type','tty','communication','baudrate','address','title','show_plot','use_cache'];
    $self->{'_legacy_tags'} = [];
    $self->{'_desc'} = {
        'mode' => 'Device control mode',
        'tty' => 'Device communication device (ex. ttyS0)',
        'communication' => 'Device communication protocol',
        'baudrate' => 'Communication speed',
        'address' => 'Device address',
	'title' => 'Optional title',
        'use_cache'=>'Determines if persisten caching of read values is allowed',
    };
    $self->{'_default'} = {
        'mode' => 'Automatic',
	'communication' => 'RS485',
	'baudrate' => 4800,
        'tty' => 'ttyS0',
        'address' => 'Z',
        'use_cache'=>'No',
    };
    $self->{'_values_type'} = [(RFC::Device::list_Multiplexer())];
    $self->{'_values_mode'} = [('Automatic')];
    $self->{'_values_communication'} = [('RS232','RS485')];
    $self->{'_values_baudrate'} = [(4800,9600,19200,38400,57600,115200)];
    $self->{'_selected'} = 0;
    bless $self,$class;
    $self->load_conf();
    $self->{'_type'} = 'VICI';
    $self->{'_set_fn_names'} = ['park']; # List of function names which can trigger a notify event
    $self->{'_ui_fn_names'} = ['park']; # List of function names which can be called from the device setup UI 
    $self->{'_ui_fn_args'} = {}; # Hash with number of args for function listed in _ui_fn_names
    $self->{'_ui_fn_desc'} = {
        'park' => 'Parks the selector in the reference position for exchanging valve installation.',
    };
    return $self->init;
}

sub park
{
    my $self = shift;
    $self->comm('AL');
    return $self->comm('NP');
}

sub Accept
{
    my $self = shift;
    my $visitor= shift;
    return undef if ($self->{'_seen_visitors'}{$visitor});
    $visitor->VisitSimple($self);
    $self->{'_seen_visitors'}{$visitor} = 1;
    return undef;
}


sub gas_all()
{
    my $self = shift;
    my $data = undef;
    if (@_)
    {
        $data = $_[0];
    }
    else
    {
        $data = RFC::Cache->new();
        $data->debug($self->debug()) if ($self->debug);
    }
    $self->check_cache($data);
    my $select = $self->channel();
    my @gas = ();
    foreach my $g (@{$self->{'_gasses'}})
    {
	next unless ($self->{'_channels'}{$g} == $select);
	my $gdev = $self->rig->get_device('gas',$g);
	push @gas, $gdev->gas() if ($gdev);
    }
    unless (@gas)
    {
	my $str = 'Warning: No gas was engaged for multiplexer '.$self->name.' for rig '.$self->rig->name;
        $self->rig->warning_mail($self,$str);
	$self->writedebug($str);
	return undef;
    }
    if (@gas > 1)
    {
	my $str = 'Alert! More than one gas is selected for multiplexer '.$self->name.' for rig ' . $self->rig->name . '. Selected gasses: '.join(', ',@gas);
        $self->rig->warning_mail($self,$str);
	$self->writedebug($str);
	return undef;
    }
    return @gas;
}


sub gas_all_name()
{
    my $self = shift;
    my $data = undef;
    if (@_)
    {
        $data = $_[0];
    }
    else
    {
        $data = RFC::Cache->new();
        $data->debug($self->debug()) if ($self->debug);
    }
    $self->check_cache($data);
    my @gas = ();
    my $select = $self->channel();
    foreach my $g (@{$self->{'_gasses'}})
    {
	next unless ($self->{'_channels'}{$g} == $select);
	my $gdev = $self->rig->get_device('gas',$g);
	push @gas, $gdev->name() if ($gdev);
    }
    unless (@gas)
    {
	my $str = 'Warning: No gas was engaged for multiplexer '.$self->name.' for rig '.$self->rig->name;
        $self->rig->warning_mail($self,$str);
	$self->writedebug($str);
	return undef;
    }
    if (@gas > 1)
    {
	my $str = 'Alert! More than one gas is selected for multiplexer '.$self->name.' for rig ' . $self->rig->name . '. Selected gasses: '.join(', ',@gas);
        $self->rig->warning_mail($self,$str);
	$self->writedebug($str);
	return undef;
    }
    return @gas;
}


sub readstring()
{
    my $self = shift;
    return undef if ($self->readstring_ignore);
    my $data = undef;
    if (@_)
    {
        $data = $_[0];
    }
    else
    {
        $data = RFC::Cache->new();
        $data->debug($self->debug()) if ($self->debug);
    }
    $self->check_cache($data);
    my $res = undef;
    my @t = $self->get_time();
    my $t = join(':',@t);
    my @gas = $self->gas_all_name($data);
    my $gas = join(',',@gas);
    $gas = 'NONE' unless (@gas);
    return join(' ',$self->name.'_selected_gas',$t,time,$gas);
}

sub gas_name
{
    my $self = shift;
    my $select = $self->channel();
    foreach my $g (@{$self->{'_gasses'}})
    {
	return $g if ($self->{'_channels'}{$g} == $select);
    }
    return undef;
}

sub read_nocache
{
    my $self = shift;
    my $data = undef;
    if (@_)
    {
        $data = $_[0];
    }
    else
    {
        $data = RFC::Cache->new();
        $data->debug($self->debug()) if ($self->debug);
    }
    $self->check_cache($data);
    my $gas = $self->gas_name();
    unless (exists($self->{'_channels'}{$gas}))
    {
	$self->writedebug("Warning, None of the gasses for multiplexer ".$self->name." was found matching gas $gas for rig ".$self->rig->name);
	return undef;
    }
    my $str = time .' '.$self->name. ' '.$gas;
    $self->rig->kanaldata($str) unless ($self->readstring_ignore);
    $self->set_cache($gas);
    $self->{'_last_read'} = Time::HiRes::time;
    return $gas;
}

sub set()
{
    my $self = shift;
    my $gas = shift;
    $self->writedebug("Setting multiplexer ".$self->name." to $gas") if ($self->debug > 2);
    unless (exists($self->{'_channels'}{$gas}))
    {
	$self->die("Error: None of the gasses for multiplexer ".$self->name." was found matching gas $gas for rig ".$self->rig->name);
	return undef;
    }
    my $data = RFC::Cache->new();
    $self->{'_selected'} = 0;
    $self->comm(sprintf('GO%02i',$self->{'_channels'}{$gas}));
    my $res = $self->channel();
    unless ($res == $self->{'_channels'}{$gas})
    {
	$self->errorlog("Error: Multiplexer ".$self->name." did not set corect port when changing gas!",$_[0]);
    }
    $self->rig->change_config_value('gas_'.$gas,'mode','Automatic');
    ### Set gas flows for not engaged gasses to 0 and mark as manual ###
    $self->writedebug("Setting non engaged gas flows to 0") if ($self->debug > 2);
    foreach my $g (@{$self->{'_gasses'}})
    {
	next if ($g eq $gas);
	$self->rig->change_config_value('gas_'.$g,'mode','Manual');
	my $dev = $self->rig->get_device('gas',$g);
	if ($dev)
	{
	    $dev->setflow(0,@_);
	}
	else
	{
	    $self->errorlog("Error: could not get device for gas $gas when changing multiplexer ".$self->name,$_[0]);
	}
    }
    ### Set gas for controler ###
    unless ($self->{'_controler'})
    {
	foreach my $d ($self->rig->list_devices('MFC'))
	{
	    my $dev = $self->rig->get_device('MFC',$d);
	    $self->{'_controler'} = $d if ($dev && $dev->multiplexer == $self->name);
	}
    }
    $self->rig->change_config_value('controler_'.$self->{'_controler'}->name,'gas',$gas) if ($self->{'_controler'});
    ### Reset error lock ###
    $self->rig->unset_warning($self);
    $self->notify('set',$gas,@_);
}


sub comm
{
    my $self = shift;
    my $cmd = shift;
    my $tty = $self->{'_tty'};
    my $adr = $self->{'_address'};
    my $mode = $self->{'_communication'};
    my $baud = $self->{'_baudrate'};
    my $cmdstr = '/'.$adr.$cmd;
    $cmdstr = $cmd if ($mode && $mode =~ /232/);
    my $res = SocketClient::serial_client_raw($self->{'_tty'},'raw',$baud,'none',8,1,$cmdstr);
    unless (defined($res) && $res =~ /\w/)
    {
	select(undef,undef,undef,0.5);
	$res = SocketClient::serial_client_raw($self->{'_tty'},'raw',$baud,'none',8,1,$cmdstr);
    }
    return $res;
}

sub get_channels
{
    my $self = shift;
    my $res = $self->comm('NP');
    if ($res && $res =~ /NP(\d+)/)
    {
	return $1+0;
    }
    return 0;
}

sub test
{
    my $self = shift;
    my $res = $self->comm('VR');
    return 1 if ($res && $res =~ /\w+/);
    return 0;
}

sub test_string()
{
    my $self = shift;
    my $res = $self->comm('VR');
    chomp $res;
    my @chs = ();
    my $select = $self->channel();
    foreach my $g (@{$self->{'_gasses'}})
    {
	my $is_select = 0;
	$is_select = 1 if (defined($self->{'_channels'}{$g}) && defined($select) && $self->{'_channels'}{$g} == $select);
	push @chs,$g.'->'.$is_select;	
    }    
    return $res."\nchannels: ".join(',',@chs);
}



sub init
{
    my $self = shift;
    $self->{'_gasses'} = [];
    $self->{'_channels'} = {};
    $self->{'_gasses'} = [$self->{'_controler'}->gasses] if ($self->{'_controler'});
    foreach my $gas (@{$self->{'_gasses'}})
    {
	my $ch = $self->get_conf($gas.'_port');
	$self->{'_channels'}{$gas} = $ch;
    }
    return $self;
}

sub print_config()
{
    my $self = shift;
    my @arr = ();
    push @arr,'Name: '.$self->name;
    push @arr,'Mode: '.$self->mode;
    push @arr,'Type: '.$self->type;
    push @arr,'TTY: '.$self->{'_tty'};
    push @arr,'Communication: '.$self->{'_communication'};
    push @arr,'Baudrate: '.$self->{'_baudrate'};
    push @arr,'Address: '.$self->{'_address'};
    my @chs = ();
    my $select = $self->channel();
    foreach my $g (@{$self->{'_gasses'}})
    {
	my $is_select = 0;
	$is_select = 1 if (defined($self->{'_channels'}{$g}) && defined($select) && $self->{'_channels'}{$g} == $select);
	push @chs,$g.'->'.$is_select;
    }    
    push @arr,'channels: '.join(',',@chs);
    return join (', ',@arr);
}


sub monitor()
{
    my $self = shift;
    my $data = undef;
    if (@_)
    {
        $data = $_[0];
    }
    my @arr = ();
    my $select = $self->channel();
    foreach my $g (@{$self->{'_gasses'}})
    {
	my $is_select = 0;
	$is_select = 1 if ($self->{'_channels'}{$g} == $select);
	push @arr,$g.' => '.$is_select;	
    }    
    return join(', ',@arr);
}


sub channel($)
{
    my $self = shift;
    my $gas = shift;
    return $self->{'_selected'} if ($self->{'_selected'});
    my $res = $self->comm('CP');
    if ($res && $res =~ /CP(\d+)/)
    {
	$self->{'_selected'} = $1+0;
    }
    return $self->{'_selected'};
}


return 1;
